package org.netkernel.module.standard;

import java.io.File;
import java.net.URI;
import java.util.*;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.container.ISpaceListener;
import org.netkernel.layer0.boot.BootUtils;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.meta.impl.ModuleMetaImpl;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.SpaceClassLoader;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.urii.IVersion;
import org.netkernel.urii.impl.Version;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class StandardModule implements IModule
{
	public static final String MODULE_DEFN="module.xml";
	public static final String STANDARD_METADATA_PARAMETERS_VALIDATION="name()!='doc' and name()!='description' and name()!='icon' and name()!='name'";
	
	private static final PairList VALIDATION = new PairList(5);
	static
	{
		VALIDATION.put("/*[name()!='module']", "MSG_STD_P_1");
		VALIDATION.put("/module/*[name()!='meta' and name()!='system' and name()!='rootspace']", "MSG_STD_P_2");
		VALIDATION.put("/module/meta/*[name()!='identity' and name()!='info']", "MSG_STD_P_3");
		VALIDATION.put("/module/meta/identity/*[name()!='uri' and name()!='version']", "MSG_STD_P_4");
		VALIDATION.put("/module/meta/info/*[name()!='name' and name()!='description' and name()!='icon' and name()!='documentation']", "MSG_STD_P_5");
		VALIDATION.put("/module/system/*[name()!='dynamic' and name()!='classloader']", "MSG_STD_P_6");
		VALIDATION.put("/module/system/classloader/*[name()!='exports' ]", "MSG_STD_P_7");
		VALIDATION.put("/module/system/classloader/exports/*[name()!='match' ]", "MSG_STD_P_8");
	}
	
	private final URI mSourceURI;
	private final IKernel mKernel;
	private final XMLReadable mDefn;
	private final File[] mToCheckForExpiry;
	private final String mModuleId;
	private final IModuleMeta mModuleMeta;
	private final SpaceClassLoader mModuleClassLoader;
	private final PairList mRootSpaces;
	private final ImportDefinition[] mImportDefinitions;
	
	public StandardModule(URI aSourceURI, XMLReadable aDefn, IKernel aKernel) throws Exception
	{	mSourceURI=aSourceURI;
		mDefn=aDefn;
		mKernel=aKernel;
		
		//validate schema outside of the space definitions
		Element root=((Document)aDefn.getContextNode()).getDocumentElement();
		PairList validation=FastSchematron.validate(root, VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aKernel.getLogger());			
			throw Utils.createFormattedException("EX_STD_CONSTRUCT", "MSG_RAW", aKernel.getLogger(), m);
		}
		
		String loc=mSourceURI.toString();
		if (loc.startsWith("file") && (!loc.endsWith(".jar")) && aDefn.getNodes("/module/system/dynamic").size()>0)
		{	File m = new File(mSourceURI);
			List<File> files=BootUtils.getJavaClasses(m);
			files.add(new File(m,MODULE_DEFN));
			files.add(new File(m,"etc/messages.properties"));
			mToCheckForExpiry=new File[files.size()];
			files.toArray(mToCheckForExpiry);
		}
		else
		{	mToCheckForExpiry=null;
		}
		
		 //build module classloader containing all classes in this module
        ClassLoader parent = this.getClass().getClassLoader();
        List exports = aDefn.getTexts("/module/system/classloader/exports/match",true);
        if (exports.size()==0) exports=null;
        mModuleClassLoader = BootUtils.getClassLoaderFrom(mSourceURI,parent,"lib",exports,aSourceURI.toString(),aKernel.getConfiguration(),aKernel.getLogger());

        //build meta
		String name=aDefn.getText("/module/meta/info/name").trim();
		if (name.length()==0)
		{	name="unnamed";
		}
		String desc=aDefn.getText("/module/meta/info/description").trim();
		String versionString=aDefn.getText("/module/meta/identity/version").trim();
		String iconString=aDefn.getText("/module/meta/info/icon").trim();
		IIdentifier icon=iconString.length()>0?new SimpleIdentifierImpl(iconString):null;
		String docString=aDefn.getText("/module/meta/info/documentation").trim();
		IIdentifier doc=docString.length()>0?new SimpleIdentifierImpl(docString):null;
		mModuleId=aDefn.getTrimText("/module/meta/identity/uri");
		mModuleMeta = new ModuleMetaImpl(name, desc, icon, doc, mSourceURI,versionString, mModuleId);
		IVersion version = new Version(versionString);
		
		
		List rootSpaceNodes = mDefn.getNodes("/module/rootspace");
		int rootSpaceCount = rootSpaceNodes.size();
		mRootSpaces = new PairList(rootSpaceCount);
		try
		{
			for (int i=0; i<rootSpaceCount; i++ )
			{
				Element rootSpaceNode = (Element)rootSpaceNodes.get(i);
				String uriString=rootSpaceNode.getAttribute("uri");
				if (uriString.length()==0)
				{	uriString=mModuleId;
					if (i>0)
					{	uriString+=':'+Integer.toString(i+1);
					}
				}
				
				IIdentifier spaceId=new SimpleIdentifierImpl(uriString);
				
				boolean isPublic=!rootSpaceNode.getAttribute("public").equals("false");
				RootSpaceWithIdentity rootSpace = new RootSpaceWithIdentity(spaceId, version, this, isPublic, rootSpaceNode);
				if (isPublic)
				{	mKernel.register(rootSpace);
				}
				mRootSpaces.put(rootSpaceNode,rootSpace);
			}
		}
		catch (Exception e)
		{
			for (int i=0; i<mRootSpaces.size(); i++)
			{	ISpaceWithIdentity s=(ISpaceWithIdentity)mRootSpaces.getValue2(i);
				mKernel.deregister(s);
			}
			throw e;
		}
		
		//find/parse all imports
		List<Node> importNodes = mDefn.getNodes("/module/rootspace//import");
		List<ImportDefinition> importDefs = new ArrayList<ImportDefinition>();
		for (Node importNode : importNodes)
		{	ImportDefinition impDef = new ImportDefinition(importNode);
			importDefs.add(impDef);
		}
		mImportDefinitions = new ImportDefinition[importDefs.size()];
		importDefs.toArray(mImportDefinitions);
		
		
		
		
	}
	
	public SpaceClassLoader getClassLoader()
	{	return mModuleClassLoader;
	}

	public void linkModule(IKernel aKernel) throws Exception
	{	linkClassLoader();
		for (int i=0; i<mRootSpaces.size(); i++)
		{	RootSpaceWithIdentity rootSpace = (RootSpaceWithIdentity)mRootSpaces.getValue2(i);
			try
			{	rootSpace.onLinkSpace(aKernel);
			}
			catch (Exception e)
			{	if (rootSpace.isPublic())
				{	mKernel.deregister(rootSpace);
				}
				mRootSpaces.removeValues(i);
				i--;
				ILogger l=aKernel.getLogger();
				l.log(ILogger.WARNING, this, "MSG_STD_MOD_EX_LINK", Utils.throwableToString(e));
			}
		}		
	}

	public void commissionModule(IKernel aKernel) throws Exception
	{	for (int i=0; i<mRootSpaces.size(); i++)
		{	Element rootSpaceNode = (Element)mRootSpaces.getValue1(i);
			RootSpaceWithIdentity rootSpace = (RootSpaceWithIdentity)mRootSpaces.getValue2(i);
			try
			{	rootSpace.onCommissionSpace(aKernel);
			}
			catch (Exception e)
			{	if (rootSpace.isPublic())
				{	mKernel.deregister(rootSpace);
				}
				mRootSpaces.removeValues(i);
				i--;
				ILogger l=aKernel.getLogger();
				l.log(ILogger.WARNING, this, "MSG_STD_MOD_EX_COMMISSION", mModuleId, Utils.throwableToString(e));
			}			
		}				
	}

	public void postCommissionModule() throws Exception
	{	for (int i=0; i<mRootSpaces.size(); i++)
		{	RootSpace rootSpace = (RootSpace)mRootSpaces.getValue2(i);
			try
			{	rootSpace.postCommissionSpace();
			}
			catch (Exception e)
			{	ILogger l=mKernel.getLogger();
				l.log(ILogger.WARNING, this, "MSG_STD_MOD_EX_POSTCOMMISSION", mModuleId, Utils.throwableToString(e));
			}
		}
	}

	public void preDecommissionModule() throws Exception
	{	mModuleClassLoader.invalidate();
		for (int i=0; i<mRootSpaces.size(); i++)
		{	RootSpace rootSpace = (RootSpace)mRootSpaces.getValue2(i);
			try
			{	rootSpace.preDecommissionSpace();
			}
			catch (Exception e)
			{	ILogger l=mKernel.getLogger();
				l.log(ILogger.WARNING, this, "MSG_STD_MOD_EX_PREDECOMMISSION", mModuleId, Utils.throwableToString(e));
			}
		}
	}

	public void decommisionModule(IKernel aKernel) throws Exception
	{	for (int i=0; i<mRootSpaces.size(); i++)
		{	RootSpaceWithIdentity rootSpace = (RootSpaceWithIdentity)mRootSpaces.getValue2(i);
			try
			{	rootSpace.onDecommissionSpace();
			}
			catch (Exception e)
			{	ILogger l=mKernel.getLogger();
				l.log(ILogger.WARNING, this, "MSG_STD_MOD_EX_DECOMMISSION", mModuleId, Utils.throwableToString(e));
			}
			mKernel.deregister(rootSpace);
		}
	}

	public IModuleMeta getMeta()
	{	return mModuleMeta;
	}

	public URI getSource()
	{	return mSourceURI;
	}

	public ISpaceWithIdentity[] getPublicSpaces()
	{	List<ISpace> spaces=new ArrayList<ISpace>(mRootSpaces.size());
		for (int i=0; i<mRootSpaces.size(); i++)
		{	RootSpace rootSpace = (RootSpace)mRootSpaces.getValue2(i);
			if (rootSpace.isPublic())
			{	spaces.add(rootSpace);
			}
		}
		ISpaceWithIdentity[] result = new ISpaceWithIdentity[spaces.size()];
		spaces.toArray(result);
		return result;
	}

	public ISpace[] getSpaces()
	{	ISpace[] result = new ISpace[mRootSpaces.size()];
		for (int i=0; i<mRootSpaces.size(); i++)
		{	RootSpace rootSpace = (RootSpace)mRootSpaces.getValue2(i);
			result[i]=rootSpace;
		}
		return result;
	}
	
	public ISpaceWithIdentity getSpace(IIdentifier aIdentifier)
	{	ISpaceWithIdentity result = null;
		for (int i=0; i<mRootSpaces.size(); i++)
		{	ISpaceWithIdentity space = (ISpaceWithIdentity)mRootSpaces.getValue2(i);
			if (space.getIdentifier().equals(aIdentifier))
			{	result = space;
				break;
			}
		}
		return result;
	}	

	public boolean isExpired(long aNow)
	{	boolean result;
		if (mToCheckForExpiry!=null)
		{	result = BootUtils.isModifiedSince(aNow, mToCheckForExpiry);
			if (result)
			{	mModuleClassLoader.invalidate();
			}
		}
		else
		{	result = false;
		}
		return result;
	}
	

	public boolean isExpiredOnRestartWithNewModules(List<IModule> newModules)
	{	boolean result=false;
		for (ImportDefinition impDef: mImportDefinitions)
		{	if (impDef.isClassLoaderDependency() || impDef.getSpace()==null)
			{	for (IModule module: newModules)
				{	for (ISpaceWithIdentity space : module.getPublicSpaces())
					{	if (space.getIdentifier().equals(impDef.getIdentifier()))
						{	IVersion v=space.getVersion();
							IVersion min=impDef.getMinVersion();
							IVersion max=impDef.getMaxVersion();
							if ( (min==null || min.isLessOrEqualThan(v)) && (max==null || max.isGreaterOrEqualThan(v)) )
							{	result=true;
								break;
							}
						}
					}
				}
			}
		}
		if (result)
		{	mModuleClassLoader.invalidate();
		}
		
		return result;
	}	

	private void linkClassLoader()
	{        
		//rescan all imports on our classloader
		List<ClassLoaderWithExports> importClassLoaders = new ArrayList<ClassLoaderWithExports>();
		for (ImportDefinition impDef: mImportDefinitions)
		{	ISpaceWithIdentity space=impDef.link(this, mKernel);
			if (space!=null && impDef.isPublic())
			{	ClassLoaderWithExports cl = space.getClassLoader();
				if (cl!=null && cl!=mModuleClassLoader && cl.hasExports())
				{	importClassLoaders.add(cl);
					impDef.setClassLoaderDependency();
				}
			}
		}
		ClassLoaderWithExports[] clArray = new ClassLoaderWithExports[importClassLoaders.size()];
		importClassLoaders.toArray(clArray);
		mModuleClassLoader.setImportedClassLoaders(clArray);		
	}
	
	private static class ImportDefinition
	{
		private final IIdentifier mIdentifier;
		private final IVersion mMinVersion;
		private final IVersion mMaxVersion;
		private boolean mIsClassLoaderDependency;
		private ISpaceWithIdentity mSpace;
		private boolean mPublic; 
		
		public ImportDefinition(Node aDefn)
		{	XMLReadable r = new XMLReadable(aDefn);
			mIdentifier=new SimpleIdentifierImpl(r.getText("uri").trim());
			String minVersionString = r.getText("version-min").trim();
			String maxVersionString = r.getText("version-max").trim();
			mMinVersion = (minVersionString.length()>0)?new Version(minVersionString):null;
			mMaxVersion = (maxVersionString.length()>0)?new Version(maxVersionString):null;
		}
		
		public IIdentifier getIdentifier()
		{	return mIdentifier;
		}
		
		public IVersion getMinVersion()
		{	return mMinVersion;
		}
		
		public IVersion getMaxVersion()
		{	return mMaxVersion;
		}
		
		public void setClassLoaderDependency()
		{	mIsClassLoaderDependency=true;
		}
		
		public boolean isClassLoaderDependency()
		{	return mIsClassLoaderDependency;
		}
		
		public ISpaceWithIdentity link(StandardModule aModule, IKernel aKernel)
		{	mSpace=aModule.getSpace(mIdentifier);
			if (mSpace==null)
			{	mSpace=aKernel.getSpace(mIdentifier, mMinVersion, mMaxVersion);
				if (mSpace!=null)
				{	mPublic=true;
				}
			}
			return mSpace;
		}
		
		public ISpaceWithIdentity getSpace()
		{	return mSpace;
		}
		
		public boolean isPublic()
		{	return mPublic;
		}
	}

}