package org.netkernel.module.standard.endpoint;

import org.netkernel.layer0.meta.*;
import org.netkernel.layer0.meta.impl.PrototypeParameterMetaImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFGrammarRequestReadOnlyImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestReadOnlyImpl;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

public abstract class ConfiguredOverlayImpl extends OverlayImpl
{
	public static final String PARAM_CONFIG="config";
	
	public static IPrototypeParameterMeta[] sPrototypeParameters = new IPrototypeParameterMeta[]
   {	new PrototypeParameterMetaImpl(PARAM_CONFIG,IIdentifier.class,IPrototypeParameterMeta.NO_DEFAULT),
		OverlayImpl.sPrototypeParameters[0]
  	};

	private INKFResponseReadOnly<IConfig> mConfigResponse;
	private boolean mBound;
	private boolean mBinding;
	private String mConfigRequestId;
	private static int sId;
	private final static IConfig sNullConfig=new NullConfig();
	
	public ConfiguredOverlayImpl()
	{	declareParameters(sPrototypeParameters);
		
		mConfigRequestId="active:"+this.getClass().getSimpleName()+"Config"+Integer.toString(sId++); // threadsafe as all initialisation is synchronous on one thread
	}
	
	/* This method ensures that the overlay get the request correctly parsed with the right grammar
	 * @see org.netkernel.layer0.nkf.impl.NKFEndpointImpl#createThisNKFRequest(org.netkernel.request.IRequest)
	 */
	protected NKFRequestReadOnlyImpl createThisNKFRequest(IRequest aRequest) throws Exception
	{	NKFRequestReadOnlyImpl result=null;
		IIdentifier endpointId=(IIdentifier)aRequest.getValue(IRequest.TYPE_RESOLVED_ENDPOINT_IDENTIFIER);
		if (endpointId!=null && mConfigResponse!=null)
		{	
			IMetaRepresentation meta=mConfigResponse.getRepresentation().getElementMeta(endpointId.toString());
			if (meta instanceof IEndpointMeta)
			{	result = new NKFGrammarRequestReadOnlyImpl(aRequest,((IEndpointMeta)meta).getIdentifierGrammar());
			}
		}
		if (result==null)
		{	result=super.createThisNKFRequest(aRequest);
		}
		return result;
	}		
	
	protected IConfig getConfig(INKFRequestContext aContext) throws Exception
	{	INKFResponseReadOnly<IConfig> response = mConfigResponse;
		if (response==null || response.isExpired())
		{	if (mBinding)
			{	if (response==null)
				{	return null;
				}
				else
				{	aContext.getKernelContext().declareDependency(response);
				}
			}
			else
			{	if (response!=null)
				{	IConfig config=response.getRepresentation();	
					destroyConfig(config);
				}
				
				mBinding=true;
				try
				{	response = aContext.sourceForResponse(mConfigRequestId,IConfig.class);
					mConfigResponse=response;
					mBound=true;
				}
				finally
				{	mBinding=false;
				}
			}
		}
		else
		{	aContext.getKernelContext().declareDependency(response);
		}				
		return response.getRepresentation();
	}
	
	public ISpaceElements getParts(INKFRequestContext aContext) throws Exception
	{	ISpaceElements result = SpaceElementsImpl.EMPTY;
		IConfig config=getConfig(aContext);
		if (config!=null)
		{	result=config.getElements(aContext);
		}
		else
		{	aContext.getKernelContext().declareExpiryDependency();
		}
		return result;
	}
	
	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{
		ISpace[] result = super.getSpaces(aContext);
		
		IConfig config=getConfig(aContext);
		if (config!=null)
		{	ISpace[] spaces=config.getSpaces(aContext);
			if (spaces.length>0)
			{	ISpace[] result2 = new ISpace[result.length+spaces.length];
				System.arraycopy(result, 0, result2, 0, result.length);
				System.arraycopy(spaces, 0, result2, result.length, spaces.length);
				result=result2;
			}
		}
		return result;
	}

	public void onResolve(INKFResolutionContext aContext) throws Exception
	{	
		INKFRequestReadOnly toResolve = aContext.getRequestToResolve();
		
		if (toResolve.getIdentifier().equals(mConfigRequestId))
		{	aContext.createResolutionResponse(this);
		}
		else if (!mBinding)
		{	
			// test mappings
			IConfig config = getConfig(aContext);
			if (config!=null)
			{	String resolvedElement=config.getResolvedElement(toResolve);
				if (resolvedElement!=null)
				{	aContext.createResolutionResponse(this, resolvedElement);
				}
			}
			if (!aContext.isResponseSet())
			{	Object requestorOverlay = toResolve.getHeaderValue(HEADER_REQUESTOR_OVERLAY);
				if (requestorOverlay==this)
				{	//overlay initiated request is trying to resolve outside overlay
					aContext.createResolutionTerminateResponse();
				}
			}
		}
	}
	
	public void onRequest(INKFRequestContext aContext) throws Exception
	{	INKFRequestReadOnly req = aContext.getThisRequest();
		if (req.getVerb()==INKFRequestReadOnly.VERB_META)
		{	onMeta(aContext);
		}
		else if (req.getIdentifier().equals(mConfigRequestId))
		{	IRequestScopeLevel scope=aContext.getKernelContext().getRequestScope();
			scope=RequestScopeLevelImpl.createOrphanedRootScopeLevel(getDelegateSpace(), scope);
			aContext.getKernelContext().setRequestScope(scope);
			clearInitialisationProblems();
			try
			{	IConfig config=createConfig(aContext);
				INKFResponse resp=aContext.createResponseFrom(config);
			}
			catch (Exception e)
			{	
				String msg=aContext.formatMessage("MSG_STD_OVERLAY_CONFIG_BAD", this.getClass().getName(), getSpace(), Utils.throwableToString(e));
				aContext.logRaw(INKFLocale.LEVEL_WARNING, msg);
				setInitialisationProblem(msg);
				aContext.createResponseFrom(sNullConfig);
				//throw e;
			}
			//set the response not to expire too quickly to avoid failing configs to constantly re-request
			long expiryPeriod=aContext.getKernelContext().getKernel().getConfiguration().getLong("netkernel.poll");
			//resp.setExpiry(INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT, System.currentTimeMillis()+expiryPeriod);
		}
		else
		{	IConfig config = getConfig(aContext);
			String id=aContext.getThisRequest().getResolvedElementId();
			IRequestScopeLevel scope=aContext.getKernelContext().getRequestScope();
			scope=RequestScopeLevelImpl.createNonDurableScopeLevel(getDelegateSpace(), scope);
			aContext.getKernelContext().setRequestScope(scope);
			config.onRequest(id,aContext);
		}
	}
	
	public void onMeta(INKFRequestContext aContext) throws Exception
	{	
		IConfig config=getConfig(aContext);
		if (config!=null)
		{	String identifier = aContext.getThisRequest().getIdentifier();
			IMetaRepresentation meta=config.getElementMeta(identifier);
			aContext.createResponseFrom(meta);
		}
		else
		{	// this case happens if we get an unresolved target that comes back to us
			Exception ex=Utils.createFormattedException("EX_STD_MAPPER_CONFIG", "MSG_STD_MAPPER_TARGET_CIRCULAR", aContext.getKernelContext().getKernel().getLogger(), aContext.getThisRequest().getIdentifier());
			INKFResponse resp = aContext.createResponseFrom(ex);
			resp.setNoCache();
			//resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
		}
	}	
	
	public void preDecommissionMapping(IRequestScopeLevel aStaticScope) throws Exception
	{	if (mConfigResponse!=null)
		{	destroyConfig(mConfigResponse.getRepresentation());
		}
		super.preDecommissionMapping(aStaticScope);
	}
	
	/** Called on first access and the subsequently when/if the configuration changes.
	 * @param aContext
	 * @return
	 * @throws Exception
	 */
	protected abstract IConfig createConfig(INKFRequestContext aContext) throws Exception;
	
	/** Called to perform any configuration specific cleanup. This is usually not needed.
	 * @param aConfig
	 * @throws Exception
	 */
	protected abstract void destroyConfig(IConfig aConfig) throws Exception;
	
	/** An overlay configuration representation. One of these must be generated each time
	 * the configuration changes. It contains active methods which are responsible for
	 * the resolving and relaying the request as well as the logical elements exposed.
	 * @author tab
	 */
	public static interface IConfig
	{
		/** Return the logical elements exposed by the overlay
		 * @param aContext the request context, if needed
		 * @return a space elements implementation
		 * @throws Exception it is best if an implement just relays on any error it gets
		 */
		public ISpaceElements getElements(INKFRequestContext aContext) throws Exception;
		
		/** Returns the meta data for the given element identifier
		 * @param aElementId an element identifier that will have been previously return in
		 * a call to getElements()
		 */
		public IMetaRepresentation getElementMeta(String aElementId);
		
		/** Resolves the given request to one of the overlays logical elements
		 * @param aRequest the request to resolve
		 * @return the logical element identifier of resolve element or null
		 * @throws Exception it is best if an implement just relays on any error it gets
		 */
		public String getResolvedElement(INKFRequestReadOnly aRequest) throws Exception;
		
		/** Handle a previously resolved request in an overlay specific way
		 * @param aElementId the resolved element identifier
		 * @param aContext the context in which to handle the request
		 * @throws Exception it is best if an implement just relays on any error it gets
		 */
		public void onRequest(String aElementId, INKFRequestContext aContext) throws Exception;
		
		/** Return any dynamic spaces which the overlay might inject into sub-requests and which
		 * it wants to expose in the metadata model.
		 * @param aContext
		 * @return an array of spaces, you can use IStandardEndpoint.NO_SPACES if there are none
		 * for a static zero length space array.
		 * @throws Exception it is best if an implement just relays on any error it gets
		 */
		public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception;
	}
	
	private static class NullConfig implements IConfig
	{
		private final ISpace[] mSpaces = new ISpace[0];
		
		public IMetaRepresentation getElementMeta(String elementId)
		{	return null;
		}

		public ISpaceElements getElements(INKFRequestContext context) throws Exception
		{	return SpaceElementsImpl.EMPTY;
		}

		public String getResolvedElement(INKFRequestReadOnly request) throws Exception
		{	return null;
		}

		public ISpace[] getSpaces(INKFRequestContext context) throws Exception
		{	return mSpaces;
		}

		public void onRequest(String elementId, INKFRequestContext context) throws Exception
		{
		}
	}
}